//*************************************************************************************************************
// 
// Тагильцев Геннадий aka Marks
// версия v5
//
// изменения v2:
// - добавлена функция boolean Engine_Proc (int, int)
// параметры 
// - требуемое положение газа
// - условная задержка (в циклах программы)
// возвращает
// true - при достижении требуемого значения
// изменения v3
// - убрал процедуру таймера, вся обработка в одном цикле.
// изменения v4
// - расчет углов только на основании данных акселерометра
// изменения v5
// - добавляем программатор черех i2c с помощью Arduino Uno + LCD Key shield
// - попробую другую библиотеку для MPU6050
//
//**************************************************************************************************************
#include <Servo.h>
#include <EEPROM.h>
#include <SoftwareSerial.h>
#include <Wire.h>
#include "Kalman.h"

#define Landing_Gear                // если ретрактов нет, ставим коммент на эту стоку строку

#define Button_Pin 12               // порт к которому подключена кнопка старт/стоп или перемычка
#define Port_Motor_ESC 2            // порт к которому подключен ESC

#define Motor_ESC_Max_THR 2000      // максимальные обороты двигателя
#define Motor_ESC_Normall_THR 1800  // номинальные (полетные) обороты
#define Motor_ESC_Landing_THR 1000  // посадочный режим
#define Motor_ESC_Min_THR 1000      // двигатель стоп
#define Time_Of_Calibrate 2         // время между Макс газ и мин газ при необходимости калибровке регулятора 2 секунды
#define Takeoff 30                  // время отведенное на взлет и уборку шасси, в течении этого времени процедура старта может быть отменена -  30 секунд 
#define GearUp 10                   // время время после запуска мотора до начала уборки шасси -  10 секунд 
#define InFly 180                   // время полета в секундах с момента завершения процедуры старта - сейчас стоит 180 секунд
#define Landing 30                  // время отводимое на посадку, в течении этого времени будут выпущены шасси и снижены обороты мотора до посадочных -  30 секунд 
#define delay_g 100                 // коэффициент торможения ))


#ifdef Landing_Gear
#define Port_Left_Wheel 6           // порт к которому подключена серва ретракта левого шасси
#define Port_Right_Wheel 7          // порт к которому подключена серва ретракта правого шасси
#define Left_Wheel_Down 2000
#define Left_Wheel_Up 1000
#define Right_Wheel_Down 1000
#define Right_Wheel_Up 2000
#define Up 1   
#define Down 0                      
#endif

  Servo Motor_ESC;                    // объект Servo для регулятора мотора
#ifdef Landing_Gear
  Servo Left_Wheel;                   // объект Servo для ретракта левого шасси
  Servo Right_Wheel;                  // объект Servo для ретракта правого шасси
#endif
  SoftwareSerial softSerial(2, 3);    // RX, TX>

//**************************************************************************************
// переменные определяющие ветвление программы
// проще говоря, состояние модели
//**************************************************************************************
  boolean Abort = false;                // процедура старта может быть прервана повторным нажатием кнопки 
//************************************************************************************
// блок объявления переменных для процедуры кальман
//************************************************************************************

Kalman kalmanX;
Kalman kalmanY;

uint8_t IMUAddress = 0x68;

/* IMU Data */
int16_t accX;
int16_t accY;
int16_t accZ;
int16_t tempRaw;
int16_t gyroX;
int16_t gyroY;
int16_t gyroZ;

double accXangle; // Angle calculate using the accelerometer
double accYangle;
double temp;
double gyroXangle = 180; // Angle calculate using the gyro
double gyroYangle = 180;
double compAngleX = 180; // Calculate the angle using a Kalman filter
double compAngleY = 180;
double kalAngleX; // Calculate the angle using a Kalman filter
double kalAngleY;

uint32_t timer;
uint32_t timeOut = 0;
uint8_t uartBufer = 0;
boolean connectFlag = false;
boolean flagTransfer = false;

//************************************************************************************
//  массив настроек программы (скоро уберу)
//************************************************************************************
  uint16_t stack[14] 
  {
  Motor_ESC_Min_THR,          // 0 минимальные обороты мотора
  Motor_ESC_Landing_THR,      // 1 нормальные обороты мотора
  Motor_ESC_Normall_THR,      // 2 нормальные обороты мотора
  Motor_ESC_Max_THR,          // 3 максимальные обороты мотора
  Time_Of_Calibrate,          // 4 время для калибровки
  Takeoff,                    // 5 время на "убежать к ручке"
  GearUp,                     // 6 момент уборки шасси с момента старта
  InFly,                      // 7 общее время полета с момента старта двигателя
  Landing,                    // 8 время на посадку, сколько времени посадочные обороты
  0,                          // 9
  0,                          // 10
  0,                          // 11
  0,                          // булево 12
  0                           // булево 13
  };

uint8_t *adrStack = (uint8_t*) stack;
uint8_t chkSumm = 0;
uint8_t i=0;
  
//************************************************************************************
// временные переменные 
//************************************************************************************
  unsigned long Current_Time;         // текущее время, мало ли где то пригодится
  unsigned long Start_Time;           // время старта по таймеру (т.е. момент нажатия на кнопку)
  unsigned long Takeoff_Time;         // время начала полета
  unsigned long GearUp_Time;          // время начала уборки шасси (т.е. с момента запуска двигателя)
  unsigned long Fly_Time;             // время завершения полета с момента запуска двигателя
  unsigned long Landing_Time;         // время останова двигателя
//*************************************************************************************

//*************************************************************************************
// эти переменные нужны что бы не задаваться вопросом "А как у нас работают сервы"
//*************************************************************************************
  int Step_Left;
  int Step_Right;
//*************************************************************************************
const int ledPin = LED_BUILTIN;  // the number of the LED pin


//*************************************************************************************
// инициализация таймера
//*************************************************************************************
void setup() 
{
  pinMode(Button_Pin, INPUT);                           // привязали кнопку старт к порту
  digitalWrite(Button_Pin, HIGH);                       // подтянули кнопку к +5V
  pinMode(ledPin, OUTPUT);
  digitalWrite(ledPin, LOW);                       // подтянули кнопку к +5V
  Motor_ESC.attach(Port_Motor_ESC,Motor_ESC_Min_THR,Motor_ESC_Max_THR); // привязали регулятор к каналу
  Wire.begin();  
  Serial.begin(9600);

//**********************************************************************************
// читаем настройки программы и
// идем в режим программирования
//**********************************************************************************
  readEEPROM();
  progMode();

//**********************************************************************************
// калибруем регулятор
//**********************************************************************************
  Motor_ESC.writeMicroseconds(Motor_ESC_Max_THR);
  delay (Time_Of_Calibrate);                                                   // ждем 2 секунды 
  Motor_ESC.writeMicroseconds(Motor_ESC_Min_THR);
  delay (500);


//**********************************************************************************

//************************************************************************************
  int i = Left_Wheel_Down-Left_Wheel_Up;
  if (i>0) {
    Step_Left = -1;
  }
  else {
    Step_Left = 1;
  }
  i = Right_Wheel_Down-Right_Wheel_Up;
  if (i>0) {
    Step_Right = -1;
  }
  else {
    Step_Right = 1;
  }
/*
  при подключении питания сервы встанут в крайнее положение, соответствующее Down
*/
#ifdef Landing_Gear
    Left_Wheel.attach(Port_Left_Wheel);      
    delay(500);
    Right_Wheel.attach(Port_Right_Wheel);    
    delay(500);
    while (!Gear_Proc(Down));
#endif
//**********************************************************************************
// вышли из режима программирования
//*********************************************************************************
  while (digitalRead(Button_Pin) == HIGH) {
  }
  delay (500);
  Start_Time = millis();
  Takeoff_Time = Start_Time + stack[5]*1000;    
  GearUp_Time = Takeoff_Time + stack[6]*1000;
  Fly_Time = Takeoff_Time + stack[7]*1000;
  Landing_Time = Fly_Time + stack[8]*1000;
//*********************************************************************************


//**********************************************************************************
// блок setup для kalman
//**********************************************************************************
  Serial.begin(115200);
  i2cWrite(0x6B,0x00); // Disable sleep mode  
  if(i2cRead(0x75,1)[0] != 0x68)                  // Read "WHO_AM_I" register
  {
    Serial.print(F("MPU-6050 with address 0x"));
    Serial.print(IMUAddress,HEX);
    Serial.println(F(" is not connected"));
    while(1);
  }      
  kalmanX.setAngle(180); // Set starting angle
  kalmanY.setAngle(180);
  timer = micros();
}

// **********************************************************************************
// основной цикл программы
// **********************************************************************************
void loop() {
  while (millis() < Takeoff_Time) Engine_Proc(stack[2],10);
  
  while (millis() < GearUp_Time) Gear_Proc(Up);
  
 while (millis() < Fly_Time) {
   int currentTHR = Motor_ESC.readMicroseconds();
   Angel();
    if (accXangle>190 &&  accXangle<350) {
      if (currentTHR != stack[3]) {
        Motor_ESC.writeMicroseconds(stack[3]);
      }
    }
    else if (accXangle<170 &&  accXangle>10) {
      if (currentTHR != stack[1]) {
        Motor_ESC.writeMicroseconds(stack[1]);
      }
    }
    else {
      Motor_ESC.writeMicroseconds(stack[2]);
    }
  }
  
  while(millis() < Landing_Time) {
    if (Engine_Proc(stack[1],10)) Gear_Proc(Down);
  }
  
  while (millis() >= Landing_Time) Engine_Proc(stack[0],10) ;
    
}

//*************************************************************************************
// процедура уборки/выпуска шасси 
// результат 
// - TRUE - завершено
// - FALSE - не завершена
// параметр - int - 
// - 0 - выпускам
// - 1 - убираем        
//*************************************************************************************
boolean Gear_Proc(int l) {
  static int i = delay_g;
  int j = Left_Wheel.readMicroseconds();
  int k = Right_Wheel.readMicroseconds();
if (l == 0) {          
// шасси выпускаем
  if (j == Left_Wheel_Down && k == Right_Wheel_Down) return true;

  if (i==0) {
    if (j > Left_Wheel_Down) j--;
    else j++;
    Left_Wheel.writeMicroseconds(j);
  }
    
  if (i==0) {
    if (k > Right_Wheel_Down) k--;
    else k++;
    Right_Wheel.writeMicroseconds(k);
  }
  
  if (!i--) i=delay_g;
}
else {
  // шасси выпускаем
  if (j == Left_Wheel_Up && k == Right_Wheel_Up) return true;

  if (i==0) {
    if (j > Left_Wheel_Up) j--;
    else j++;
    Left_Wheel.writeMicroseconds(j);
  }
    
  if (i==0) {
    if (k > Right_Wheel_Up) k--;
    else k++;
    Right_Wheel.writeMicroseconds(k);
 }
  if (!i--) i=delay_g;
}
  return false;
}


//*************************************************************************************
// Процедура работы с регуляторм
// возвращает true когда результат достигнут
// первый параметр - int - значение до которого надо изменить газ
// второй параметр - int - задержка
//*************************************************************************************
boolean Engine_Proc(int j1, int i1) {
  static int i = 0;
  int j =  Motor_ESC.readMicroseconds();

  if (j==j1) return true;
  
  if (i == i1 && j<j1) {
    j++;   
    Motor_ESC.writeMicroseconds(j);
  }
  if (i == i1 && j>j1) {
    j--;   
    Motor_ESC.writeMicroseconds(j);
  }
  if (!i--) i=i1;
  if (digitalRead(Button_Pin) == LOW) Abort = true ;  
  return false;  
}


//***************************************************************************************
// кальман для гироскопа
// считает углы ))
//***************************************************************************************
void Angel() {
  uint8_t* data = i2cRead(0x3B,14);  
  accX = ((data[0] << 8) | data[1]);
  accY = ((data[2] << 8) | data[3]);
  accZ = ((data[4] << 8) | data[5]);  
  /* Calculate the angls based on the different sensors and algorithm */
  accYangle = (atan2(accX,accZ)+PI)*RAD_TO_DEG;
  accXangle = (atan2(accY,accZ)+PI)*RAD_TO_DEG;    
  delay (1);
}


//***************************************************************************************
// запись i2c
//***************************************************************************************

void i2cWrite(uint8_t registerAddress, uint8_t data){
  Wire.beginTransmission(IMUAddress);
  Wire.write(registerAddress);
  Wire.write(data);
  Wire.endTransmission(); // Send stop
}

//***************************************************************************************
// чтение i2c
//***************************************************************************************

uint8_t* i2cRead(uint8_t registerAddress, uint8_t nbytes) {
  uint8_t data[nbytes];  
  Wire.beginTransmission(IMUAddress);
  Wire.write(registerAddress);
  Wire.endTransmission(false); // Don't release the bus
  Wire.requestFrom(IMUAddress, nbytes); // Send a repeated start and then release the bus after reading
  for(uint8_t i = 0; i < nbytes; i++)
    data[i] = Wire.read();
  return data;
} 

//*************************************************************************************
// чтение EEPROM
//*************************************************************************************
void readEEPROM()
{
  chkSumm = 0;
  int i = 0;
  if (EEPROM[0] == 0xFF)
  {
    for (i=0; i <= 27; i++) EEPROM[i] = adrStack[i];
  }
  
  for (i=0; i <= 27; i++)
  {
    adrStack[i] = EEPROM[i];
    chkSumm = chkSumm + adrStack[i];
  } 
}

//***************************************************************************
// прием данных от программатора
//***************************************************************************
void resiveSetting() 
{
  chkSumm = 0;
  timer = millis();
  i=0;
  uartBufer = 0;
  while (1)
  {
    if (millis() - timer > 500) 
    {
      digitalWrite(ledPin, HIGH);                       // данные не приняты по таймауту
      return;
    }
    if (Serial.available() > 0) 
    {
      uartBufer = Serial.read();
      adrStack[i] = uartBufer;
      chkSumm = chkSumm + uartBufer;
      i++;
      if (i == 28) break; //  return;
    } 
  }
  while (1)
  {
    if (millis() - timer > 500) 
    {
      digitalWrite(ledPin, HIGH);                       // зажгли светодиод
      return;
    }
    if (Serial.available() > 0)                         // ждем контрольную сумму
    {
      uartBufer = Serial.read();
      if (uartBufer == chkSumm) 
      {
        digitalWrite(ledPin, LOW);                      // погасили светодиод
        for (i=0; i<=27; i++) EEPROM[i] = adrStack[i];
//        Serial.write (0x99);
        return;
      }
    } 
  }
}

//***************************************************************************
// передача данных в программатор
//***************************************************************************

void sendSetting()
{
  chkSumm = 0;
  for (int i = 0; i <= 27; i++)
  {
    Serial.write(adrStack[i]);                  // отправили байт
    chkSumm = chkSumm + adrStack[i];            // считаем контрольную сумму
  }
  Serial.write(chkSumm);
}
//***********************************************************************************
// режим программирования
//***********************************************************************************
void progMode () 
{
  i = 0;  
  timeOut = millis();
  connectFlag = false;
  uartBufer = 0;
  while (1)
  {
    if (millis()-timeOut > 3000 && !connectFlag)        // блок таймаута при включении
    {
      digitalWrite(ledPin, HIGH);                       // подтянули кнопку к +5V
      return;                                           // случка программатора и таймера не состоялась в течении 3  секунд после включения таймера
    }
    if (Serial.available()>0)                           // ждем данные из UART
    {
      uartBufer = Serial.read();
      switch (uartBufer)
      {
      case 0x55:                                        // запрос соединения
        connectFlag = true;
        Serial.write (0x55);
        break;
      case 0x5A:                                        // запрос настроек
        sendSetting();                                  // отправляем настройки
        break;
      case 0xA5:                                        // передача настроек от программатора
        resiveSetting();                                // принимаем настройки
      break;
      default:
      break;
      }
    }
  }
}
//****************************************************************************************